#include <string.h>
extern "C"
{
#include "doomdef.h"
#include "doomtype.h"

#include "w_wad.h"
//#include "templates.h"
//#include "c_cvars.h"
#include "i_system.h"
}
#include "fmopl.h"
#include "opl_mus_player.h"

static bool  opl_onechip = true;
static int opl_frequency = 11025;
static OPLmusicBlock* Music = NULL;
static unsigned char* NewMusicData = NULL;
static int NewMusicDataLen = 0;

static unsigned char* CurrentMusicData = NULL;
static int CurrentMusicDataLen = 0;

static bool gotInstrs;
static int sample_count = 512;
static unsigned char* IntrumentBank = NULL;
static int InstrumentBankLen = 0;
static int MusicVolume = 15;
static int MusicQuality = 0;

extern "C" void InitOPLMusic()
{
	if (!gotInstrs)
	{
		int lumpnum = W_GetNumForName("GENMIDI");
		InstrumentBankLen = W_LumpLength(lumpnum);
		IntrumentBank = (unsigned char*) malloc(InstrumentBankLen);
		W_ReadLump(lumpnum, IntrumentBank);

		 
		gotInstrs = true;
	}
}

extern "C" void CreateOPLMusic(char* aData, int aLen)
{		
	unsigned char* musicData = (unsigned char*) malloc(aLen);
	NewMusicDataLen = aLen;
	memcpy(musicData, aData, aLen);
	NewMusicData = musicData;
}

extern"C" void StopOPLSong()
{		
	if(Music != NULL)
	{
		OPLmusicBlock* musicBlock = Music;
		
		musicBlock->iChipAccess.Wait();
		Music = NULL;
		musicBlock->iChipAccess.Signal();
		delete musicBlock;
	}
}

extern"C" void DestroyOPLMusic()
{
	StopOPLSong();
	free(CurrentMusicData);
	CurrentMusicData = NULL;
	CurrentMusicDataLen = 0;
	free(IntrumentBank);
	IntrumentBank = NULL;
	gotInstrs = false;	
}

extern "C" void SetOPLMusicVolume(int aVolume)
{
	MusicVolume = aVolume;
}

extern "C" void SetOPLMusicQuality(int aQuality)
{
	MusicQuality = aQuality;
	if(aQuality)
	{
		opl_frequency = 22050;
		sample_count = 1024;
	}
	else
	{
		opl_frequency= 11025;
		sample_count = 512;
	}

	if(CurrentMusicData != NULL && CurrentMusicDataLen>0)
	{
		NewMusicDataLen = CurrentMusicDataLen; 
		NewMusicData = CurrentMusicData;
	}
}

extern "C" void OPLMusicMixer(char* aData, int aLen)
{
	if(NewMusicData != NULL)
	{
		int rate = opl_frequency;
		int samples = sample_count;//rate/14;
		delete Music;
		Music = NULL;

		if(CurrentMusicData != NewMusicData)
		{
			free(CurrentMusicData);
			CurrentMusicData = NULL;
			CurrentMusicDataLen = 0;
		}

		Music = new OPLmusicBlock (NewMusicData, NewMusicDataLen, rate, samples);
		Music->SetLooping(true);
	
		CurrentMusicData = NewMusicData;
		CurrentMusicDataLen = NewMusicDataLen;
		NewMusicData = NULL;
		NewMusicDataLen = 0;

		int failed = OPLloadBank (IntrumentBank, InstrumentBankLen);
		if (failed)
		{
			return;
		}
	}

	if(Music != NULL && MusicVolume>0)
	{	
		Music->ServiceStream(aData, aLen);
	}
}

OPLmusicBlock::OPLmusicBlock (unsigned char* aScoreData, int len, int rate, int maxSamples)
	: SampleRate (rate), NextTickIn (0), Looping (false)
{
	iChipAccess.CreateLocal(1);
//	InitializeCriticalSection (&ChipAccess);

	scoredata = NULL;
	SampleBuff = NULL;
	TwoChips = !opl_onechip;


	SampleBuff = new signed short int[maxSamples];

	scoredata = new BYTE[len];
	memcpy(scoredata, aScoreData, len);

	if (OPLinit (TwoChips + 1, rate))
	{
		delete[] scoredata;
		scoredata = NULL;
		return;
	}

	SamplesPerTick = ((rate/14)+5)/10;	// round to nearest, not lowest
	Restart ();
}

OPLmusicBlock::~OPLmusicBlock ()
{
	iChipAccess.Wait();
	if (scoredata != NULL)
	{
		OPLdeinit ();
		delete[] scoredata;
		scoredata = NULL;
	}
	if (SampleBuff != NULL)
	{
		delete[] SampleBuff;
		SampleBuff = NULL;
	}
	iChipAccess.Signal();
	iChipAccess.Close();
}

void OPLmusicBlock::ResetChips ()
{
	TwoChips = !opl_onechip;
	iChipAccess.Wait();
	OPLdeinit ();
	OPLinit (TwoChips + 1, SampleRate);
	iChipAccess.Signal();
}

bool OPLmusicBlock::IsValid () const
{
	return scoredata != NULL;
}

void OPLmusicBlock::SetLooping (bool loop)
{
	Looping = loop;
}

void OPLmusicBlock::Restart ()
{
	OPLstopMusic (this);
	OPLplayMusic (this);
	MLtime = 0;
	playingcount = 0;
	score = scoredata + ((MUSheader *)scoredata)->scoreStart;
}

bool OPLmusicBlock::ServiceStream (void *buff, int numbytes)
{
	signed short int *samples = (signed short int*)buff;
	signed short int *samples1 = SampleBuff;
	int numsamples = numbytes / 4;
	if(!MusicQuality)	
		numsamples=numsamples>>1;

	bool prevEnded = false;
	bool res = true;

	memset (SampleBuff, 0, numsamples*2);
	iChipAccess.Wait();
	while (numsamples > 0)
	{
		int samplesleft = min (numsamples, NextTickIn);

		if (samplesleft > 0)
		{
			YM3812UpdateOne (0, samples1, samplesleft);
			if (TwoChips)
			{
				YM3812UpdateOne (1, samples1, samplesleft);
			}
			NextTickIn -= samplesleft;
			numsamples -= samplesleft;
			samples1 += samplesleft;
		}
		
		if (NextTickIn == 0)
		{
			int next = playTick (this);
			if (next == 0)
			{ // end of song
				if (!Looping || prevEnded)
				{
					if (numsamples > 0)
					{
						YM3812UpdateOne (0, samples1, numsamples);
						if (TwoChips)
						{
							YM3812UpdateOne (1, samples1, numsamples);
						}
					}
					res = false;
					break;
				}
				else
				{
					// Avoid infinite loops from songs that do nothing but end
					prevEnded = true;
					Restart ();
				}
			}
			else
			{
				prevEnded = false;
				NextTickIn = SamplesPerTick * next;
				MLtime += next;
			}
		}
	}
	iChipAccess.Signal();
	numsamples = numbytes / 4;
	samples1 = SampleBuff;
	if(!MusicQuality)	
		numsamples=numsamples>>1;
#if defined(_MSC_VER) && defined(USEASM)
	if (CPU.bCMOV && numsamples > 1)
	{
		__asm
		{
			mov ecx, numsamples
			mov esi, samples1
			mov edi, samples
			shr ecx, 1
			lea esi, [esi+ecx*8]
			lea edi, [edi+ecx*4]
			neg ecx
			mov edx, 0x00007fff
looper:		mov eax, [esi+ecx*8]
			mov ebx, [esi+ecx*8+4]

			// Shift the samples down to reduce the chance of clipping at the high end.
			// Since most of the waveforms we produce only use the upper half of the
			// sine wave, this is a good thing. Although it does leave less room at
			// the bottom of the sine before clipping, almost none of the songs I tested
			// went below -9000.
			sub eax, 18000
			sub ebx, 18000

			// Clamp high
			cmp eax, edx
			cmovg eax, edx
			cmp ebx, edx
			cmovg ebx, edx

			// Clamp low
			not edx
			cmp eax, edx
			cmovl eax, edx
			cmp ebx, edx
			cmovl ebx, edx

			not edx
			mov [edi+ecx*4], ax
			mov [edi+ecx*4+2], bx

			inc ecx
			jnz looper

			test numsamples, 1
			jz done

			mov eax, [esi+ecx*8]
			sub eax, 18000
			cmp eax, edx
			cmovg eax, edx
			not edx
			cmp eax, edx
			cmovl eax, edx
			mov [edi+ecx*2], ax
done:
		}
	}
	else
#endif
	{
			if(MusicQuality)		
				for (int i = 0; i < numsamples; i++)
				{
					samples[(i*2)  ] = samples[(i*2)+1 ] = ((samples1[i  ])*MusicVolume)>>5;
					
				}	
			else
			{				
				for (int i = 0; i < numsamples; i++)
				{
					samples[(i*4)  ] = samples[(i*4)+1 ] = samples[(i*4) +2 ] = samples[(i*4)+3 ] = ((samples1[i  ])*MusicVolume)>>5;					
				}	
			}
	}
	return true;
}
